/* Generator is (c) James Ponder, 1997-2001 http://www.squish.net/generator/ */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>

#include "generator.h"
#include "snprintf.h"

#include "ui.h"
#include "memz80.h"
#include "cpu68k.h"
#include "mem68k.h"
#include "cpuz80.h"
#include "vdp.h"
#include "gensound.h"

#ifdef DREAMCAST
#include "gui.h"
#endif

/*** variables externed in generator.h ***/

unsigned int gen_quit = 0;
unsigned int gen_debugmode = 0;
unsigned int gen_loglevel = 3;  /* 2 = NORMAL, 1 = CRITICAL */
unsigned int gen_autodetect = 1; /* 0 = no, 1 = yes */
unsigned int gen_musiclog = 0; /* 0 = no, 1 = GYM, 2 = GNM */
t_cartinfo gen_cartinfo;
char gen_leafname[128];

static int gen_freerom = 0;

static uint8 genesis_rom[0x400000];
static uint8 smd_decode[0x4000];

/*** forward references ***/

void gen_nicetext(char *out, char *in, unsigned int size);
uint16 gen_checksum(uint8 *start, unsigned int length);
void gen_setupcartinfo(void);

/*** Program entry point ***/

extern uint8 romdisk_boot[];

pvr_init_params_t params =
{
	{PVR_BINSIZE_16, PVR_BINSIZE_0, PVR_BINSIZE_16, PVR_BINSIZE_0, PVR_BINSIZE_0, },
	512*1024
};

int main(int argc, char *argv[])
{
  int retval;
  t_sr test;

  test.sr_int = 0;
  test.sr_struct.c = 1;
  if (test.sr_int != 1) {
    fprintf(stderr, "%s: compilation variable BYTES_HIGHFIRST not set "
            "correctly\n", argv[0]);
    return 1;
  }
  kos_init_all(IRQ_ENABLE|THD_ENABLE, romdisk_boot);
#ifdef PAL_DREAMCAST
  vid_init(DM_640x480_PAL_IL, PM_RGB565);
#else
  vid_init(DM_640x480_NTSC_IL, PM_RGB565);
#endif
  pvr_init(&params);
  init_ta_text();

  /* initialise user interface */
  if ((retval = ui_init(argc, argv)))
    return retval;

  /* initialise 68k memory system */
  if ((retval = mem68k_init()))
    ui_err("Failed to initialise mem68k module (%d)", retval);

  /* initialise z80 memory system */
  if ((retval = memz80_init()))
    ui_err("Failed to initialise memz80 module (%d)", retval);

  /* initialise vdp system */
  if ((retval = vdp_init()))
    ui_err("Failed to initialise vdp module (%d)", retval);

  /* initialise cpu system */
  if ((retval = cpu68k_init()))
    ui_err("Failed to initialise cpu68k module (%d)", retval);

  /* initialise z80 cpu system */
  if ((retval = cpuz80_init()))
    ui_err("Failed to initialise cpuz80 module (%d)", retval);

  /* initialise sound system */
  if ((retval = sound_init()))
    ui_err("Failed to initialise sound module (%d)", retval);

  gui_runner();
  return 0;
}

/*** gen_reset - reset system ***/

void gen_reset(void)
{
  vdp_reset();
  cpu68k_reset();
  cpuz80_reset();
  if (sound_reset()) {
    ui_err("sound failure");
  }
}

/*** gen_softreset - reset system ***/

void gen_softreset(void)
{
  cpu68k_reset();
}
void update_progress(const uint8 *romname, uint32 total, uint32 current, uint8 *info, int abortable )
{
	pvr_wait_ready();
	pvr_scene_begin();
	pvr_list_begin(PVR_LIST_OP_POLY);
	pvr_dummy_poly(PVR_LIST_OP_POLY);
	pvr_list_finish();
	pvr_list_begin(PVR_LIST_TR_POLY);
	draw_poly_strf(40.0f, 224.0f, 100.0f, 1.0f, 1.0f, 1.0f, 1.0f, "%s %s", info, romname);
	draw_poly_strf(40.0f, 250.0f, 100.0f, 1.0f, 1.0f, 1.0f, 1.0f, "%u/%u", current, total);
	if(abortable)
	{
		draw_poly_strf(40.0f, 400.0f, 100.0f, 1.0f, 1.0f, 1.0f, 1.0f, "PRESS B TO CANCEL LOADING");
	}
	pvr_list_finish();
	pvr_scene_finish();
}

int ui_check_cond(uint32 flags);

/*** gen_loadimage - load ROM image ***/
char *gen_loadimage(const char *filename)
{
  int file, imagetype, bytes, bytesleft;
  const char *extension;
  uint8 *buffer;
  unsigned int blocks, x, i;
  char *p;

  imagetype = 1;                /* BIN file by default */

  /* Check extension is not wrong */
  extension = filename + strlen(filename) - 3;
  if (!stricmp(extension, "smd"))
  {
    LOG_VERBOSE(("attempting to recognize as SMD file\n"));
    imagetype = 2;
  }

  if ((file = fs_open(filename, O_RDONLY)) == -1)
  {
    ui_err("unable to open file.\n");
    return ("Unable to open file.");
  }
  
  cpu68k_romlen = fs_total(file);
  if(imagetype == 2)
    cpu68k_romlen -= 512;

  /* Load file */
  if (cpu68k_romlen < 0x200)
  {
    ui_err("File is too small\n");
    return ("File is too small");
  }
  
  /* is this icky? */
  if ((p = strrchr(filename, '/')) == NULL &&
      (p = strrchr(filename, '\\')) == NULL)
  {
    snprintf(gen_leafname, sizeof(gen_leafname), "%s", filename);
  }
  else
  {
    snprintf(gen_leafname, sizeof(gen_leafname), "%s", p + 1);
  }
  if ((p = strrchr(gen_leafname, '.')) != NULL)
  {
    if ((!stricmp(p, ".smd")) || (!stricmp(p, ".bin")))
      *p = '\0';
  }
  if (gen_leafname[0] == '\0')
    snprintf(gen_leafname, sizeof(gen_leafname), "rom");

  cpu68k_rom = (uint8 *)&genesis_rom;
  gen_freerom = 0;
  memset(cpu68k_rom, 0, cpu68k_romlen);
  buffer = cpu68k_rom;
  bytesleft = cpu68k_romlen;
  /* seek past header for SMD files */
  update_progress(gen_leafname, cpu68k_romlen, cpu68k_romlen-bytesleft, "Now Loading:", 0);
  if(imagetype == 2)
    fs_seek(file, 512, SEEK_SET);
  do
  {
    update_progress(gen_leafname, cpu68k_romlen, cpu68k_romlen-bytesleft, "Now Loading:", 1);
    if ((bytes = fs_read(file, buffer, 0x4000)) <= 0)
      break;
    buffer += bytes;
    bytesleft -= bytes;
  }
  while (bytesleft >= 0 && !ui_check_cond(CONT_B));
  fs_close(file);
  if(bytesleft)
  {
	  return("Rom Load Cancelled or error occured");
  }

  /* convert to standard BIN file format */

  switch (imagetype)
  {
  case 1:                      /* BIN */
    break;
  case 2:                      /* SMD */
    blocks = cpu68k_romlen / 16384;
    /* if(cpu69k_romlen % 16384) */
    if (blocks * 16384 != cpu68k_romlen)
      {
	free(cpu68k_rom);
	ui_err("Image is corrupt.");
	return ("Image is corrupt.");
      }
    for (i = 0; i < blocks; i++)
    {
      update_progress(filename, blocks, i, "Decoding SMD:", 0);
      buffer = cpu68k_rom + (i*0x4000);
      memcpy(smd_decode, buffer, 0x4000);
      for (x = 0; x < 0x2000; x++)
      {
        *buffer++ = smd_decode[x + 0x2000];
	*buffer++ = smd_decode[x];
      }
    }
    break;
  default:
    ui_err("Unknown image type");
    return ("Unknown image type");
    break;
  }

  gen_setupcartinfo();

  if (gen_autodetect)
  {
    vdp_pal = (!gen_cartinfo.flag_usa && !gen_cartinfo.flag_japan &&
               gen_cartinfo.flag_europe) ? 1 : 0;
  }

  /* reset system */
  gen_reset();

  while(!ui_check_cond(CONT_START))
  {
	pvr_wait_ready();
	pvr_scene_begin();
	pvr_list_begin(PVR_LIST_OP_POLY);
	pvr_dummy_poly(PVR_LIST_OP_POLY);
	pvr_list_finish();
	pvr_list_begin(PVR_LIST_TR_POLY);
	draw_poly_strf(40.0f, 200.0f, 100.0f, 1.0f, 1.0f, 1.0f, 1.0f, "Loaded:");
	draw_poly_strf(40.0f, 224.0f, 100.0f, 1.0f, 1.0f, 1.0f, 1.0f, "DOMESTIC: %s", gen_cartinfo.name_domestic);
	draw_poly_strf(40.0f, 250.0f, 100.0f, 1.0f, 1.0f, 1.0f, 1.0f, "OVERSEAS: %s", gen_cartinfo.name_overseas);
	draw_poly_strf(40.0f, 276.0f, 100.0f, 1.0f, 1.0f, 1.0f, 1.0f, "VER: %s CRC: %04X COUNTRY:%s", gen_cartinfo.version, gen_cartinfo.checksum, gen_cartinfo.country);
	
  	if (gen_cartinfo.checksum != (cpu68k_rom[0x18e] << 8 | cpu68k_rom[0x18f]))
	{
		draw_poly_strf(40.0f, 276.0f, 100.0f, 1.0f, 1.0f, 1.0f, 1.0f, "Warning CRC does not match in rom: %04X", (cpu68k_rom[0x18e] << 8 | cpu68k_rom[0x18f]));
	}
	draw_poly_strf(40.0f, 400.0f, 100.0f, 1.0f, 1.0f, 1.0f, 1.0f, "PRESS START");
	pvr_list_finish();
	pvr_scene_finish();
  }
  gen_quit = 0;
  return NULL;
}

/* setup gen_cartinfo from current loaded rom */
void gen_setupcartinfo(void)
{
  unsigned int i;
  char *p;

  memset(&gen_cartinfo, 0, sizeof(gen_cartinfo));
  gen_nicetext(gen_cartinfo.console, (char *)(cpu68k_rom + 0x100), 16);
  gen_nicetext(gen_cartinfo.copyright, (char *)(cpu68k_rom + 0x110), 16);
  gen_nicetext(gen_cartinfo.name_domestic, (char *)(cpu68k_rom + 0x120), 48);
  gen_nicetext(gen_cartinfo.name_overseas, (char *)(cpu68k_rom + 0x150), 48);
  if (cpu68k_rom[0x180] == 'G' && cpu68k_rom[0x181] == 'M')
  {
    gen_cartinfo.prodtype = pt_game;
  }
  else if (cpu68k_rom[0x180] == 'A' && cpu68k_rom[0x181] == 'I')
  {
    gen_cartinfo.prodtype = pt_education;
  }
  else
  {
    gen_cartinfo.prodtype = pt_unknown;
  }
  gen_nicetext(gen_cartinfo.version, (char *)(cpu68k_rom + 0x182), 12);
  gen_cartinfo.checksum = gen_checksum(((uint8 *)cpu68k_rom) + 0x200,
                                       cpu68k_romlen - 0x200);
  gen_nicetext(gen_cartinfo.memo, (char *)(cpu68k_rom + 0x1C8), 28);
  for (i = 0x1f0; i < 0x1ff; i++)
  {
    if (cpu68k_rom[i] == 'J')
      gen_cartinfo.flag_japan = 1;
    if (cpu68k_rom[i] == 'U')
      gen_cartinfo.flag_usa = 1;
    if (cpu68k_rom[i] == 'E')
      gen_cartinfo.flag_europe = 1;
  }
  if (cpu68k_rom[0x1f0] >= '1' && cpu68k_rom[0x1f0] <= '9')
  {
    gen_cartinfo.hardware = cpu68k_rom[0x1f0] - '0';
  }
  else if (cpu68k_rom[0x1f0] >= 'A' && cpu68k_rom[0x1f0] <= 'F')
  {
    gen_cartinfo.hardware = cpu68k_rom[0x1f0] - 'A' + 10;
  }
  p = gen_cartinfo.country;
  for (i = 0x1f0; i < 0x200; i++)
  {
    if (cpu68k_rom[i] != 0 && cpu68k_rom[i] != 32)
      *p++ = cpu68k_rom[i];
  }
  *p = '\0';
}

/*** get_nicetext - take a string, remove spaces and capitalise ***/

void gen_nicetext(char *out, char *in, unsigned int size)
{
  int flag, i;
  int c;
  char *start = out;

  flag = 0;                     /* set if within word, e.g. make lowercase */
  i = size;                     /* maximum number of chars in input */
  while ((c = *in++) && --i > 0)
  {
    if ((c >= 'A' && c <= 'Z') || (c >= 'a' && c <= 'z'))
    {
      if (!flag)
      {
        /* make uppercase */
        flag = 1;
        if (islower(c))
          *out++ = c - 'z' + 'Z';
        else
          *out++ = c;
      }
      else
      {
        /* make lowercase */
        if (isupper(c))
          *out++ = (c) - 'Z' + 'z';
        else
          *out++ = c;
      }
    }
    else if (c == ' ')
    {
      if (flag)
        *out++ = c;
      flag = 0;
    }
    else if (isprint(c) && c != '\t')
    {
      flag = 0;
      *out++ = c;
    }
  }
  while (out > start && out[-1] == ' ')
    out--;
  *out++ = '\0';
}

/*** gen_checksum - get Genesis-style checksum of memory block ***/

uint16 gen_checksum(uint8 *start, unsigned int length)
{
  uint16 checksum = 0;

  if (length & 1)
  {
    length &= ~1;
    LOG_CRITICAL(("checksum routines given odd length (%d)", length));
  }

  for (; length; length -= 2, start += 2)
  {
    checksum += start[0] << 8;
    checksum += start[1];
  }
  return checksum;
}

