#include <stdarg.h>
#include <stdio.h>
#include "generator.h"

/*
 * Some routines were borrowed from code by Dan Potter.
 */

pvr_ptr_t ta_text_txr = 0;
pvr_poly_hdr_t font_header;
pvr_poly_hdr_t color_header;

void init_ta_text()
{
	uint16 *txr;
	uint32 x, y;
	pvr_poly_cxt_t tmp;
	pvr_poly_cxt_t tmp2;
	
	ta_text_txr = pvr_mem_malloc(256*256*2);
	txr = ta_txr_map(ta_text_txr);
	for(y = 0; y < 8; y++)
	{
		for(x = 0; x < 16; x++)
		{
			if(x != 0 || y != 0)
			{
				bfont_draw(txr, 256, 0, (y<<4)+x);
			}
			txr += 16;
		}
		txr += 23*256;
	}
	pvr_poly_cxt_txr(&tmp, PVR_LIST_TR_POLY, PVR_TXRFMT_ARGB1555|PVR_TXRFMT_NONTWIDDLED, 256, 256, ta_text_txr, PVR_FILTER_NONE);
	pvr_poly_compile(&font_header, &tmp);
	pvr_poly_cxt_col(&tmp2, PVR_LIST_TR_POLY);
	pvr_poly_compile(&color_header, &tmp2);
}

/* Draw one font character (12x24); assumes polygon header already sent */
void draw_poly_char(float x1, float y1, float z1, float a, float r, float g, float b, int c) {
	pvr_vertex_t vert;
	int ix = (c % 16) * 16;
	int iy = (c / 16) * 24;
	float u1 = ix * 1.0f / 256.0f;
	float v1 = iy * 1.0f / 256.0f;
	float u2 = (ix+12) * 1.0f / 256.0f;
	float v2 = (iy+24) * 1.0f / 256.0f;

	vert.flags = PVR_CMD_VERTEX;
	vert.x = x1;
	vert.y = y1 + 24.0f;
	vert.z = z1;
	vert.u = u1;
	vert.v = v2;
	vert.argb = PVR_PACK_COLOR(a,r,g,b);
	vert.oargb = 0;
	pvr_prim(&vert, sizeof(vert));
	
	vert.x = x1;
	vert.y = y1;
	vert.u = u1;
	vert.v = v1;
	pvr_prim(&vert, sizeof(vert));
	
	vert.x = x1 + 12.0f;
	vert.y = y1 + 24.0f;
	vert.u = u2;
	vert.v = v2;
	pvr_prim(&vert, sizeof(vert));

	vert.flags = PVR_CMD_VERTEX_EOL;
	vert.x = x1 + 12.0f;
	vert.y = y1;
	vert.u = u2;
	vert.v = v1;
	pvr_prim(&vert, sizeof(vert));
}

/* Draw a set of textured polygons at the given depth and color that
   represent a string of text. */
static char strbuf[1024];
void draw_poly_strf(float x1, float y1, float z1, float a, float r,
		float g, float b, char *fmt, ...) {
	va_list args;
	char *s;
	
	va_start(args, fmt);
	vsprintf(strbuf, fmt, args);
	va_end(args);

	pvr_prim(&font_header, sizeof(font_header));
	s = strbuf;
	while (*s) {
		if (*s == ' ') {
			x1 += 12.0f; s++;
		} else {
			draw_poly_char(x1+=12.0f, y1, z1, a, r, g, b, *s++);
		}
	}
}

/* Draw a polygon for a shaded box; wow, a nasty looking func =) */
void draw_poly_box(float x1, float y1, float x2, float y2, float z,
		float a1, float r1, float g1, float b1,
		float a2, float r2, float g2, float b2) {
	pvr_vertex_t vert;
	
	pvr_prim(&color_header, sizeof(color_header));
	
	vert.flags = PVR_CMD_VERTEX;
	vert.x = x1; vert.y = y2; vert.z = z;
	vert.argb = PVR_PACK_COLOR((a1+a2)/2, (r1+r2/2), (g1+g2)/2, (b1+b2)/2);
	pvr_prim(&vert, sizeof(vert));
	
	vert.y = y1;
	vert.argb = PVR_PACK_COLOR(a1, r1, g1, b1);
	pvr_prim(&vert, sizeof(vert));
	
	vert.x = x2; vert.y = y2;
	vert.argb = PVR_PACK_COLOR(a2, r2, g2, b2);
	pvr_prim(&vert, sizeof(vert));
	
	vert.flags = PVR_CMD_VERTEX_EOL;
	vert.y = y1;
	vert.argb = PVR_PACK_COLOR((a1+a2)/2, (r1+r2/2), (g1+g2)/2, (b1+b2)/2);
	pvr_prim(&vert, sizeof(vert));
}

