#include "generator.h"
#include "gui.h"
#include "ui-console.h"
#include "vdp.h"
#include "gensound.h"

typedef struct
{
	char *string;
	int value;
} T_OPTION_STRING;

typedef struct
{
	char desc[80];
	uint8 *variable;
	uint8 current;
	uint8 max;
} T_OPTIONS;

extern uint8 auto_frameskip;
extern uint8 auto_frameskip_count;
extern uint8 ui_vdpsimple;
extern unsigned int cpuz80_on;

T_OPTIONS options[] =
{
	// SOUND OPTIONS
	{"sound enabled", &sound_on, 0, 1},

	// VIDEO OPTIONS
	{"Simple [cell]", &ui_vdpsimple, 1},
	{"Auto-Frameskip [experimental]", &auto_frameskip, 0, 1},
	{"Frameskip Count [experimental]", &auto_frameskip_count, 4, 10},

	// CPU OPTIONS
	{"Z80 Emulation", (uint8*)&cpuz80_on, 0, 1},

	// NULL OPTION TO SIGNAL END OF LIST
	{"\0", NULL, 1, 1},
};

void draw_options(int selected)
{
	int y = 58;
	int counter = 0;
	static float alpha = 1.0f, dalpha = -0.01f;

	while(options[counter].variable != NULL)
	{
		if(options[counter].max > 1)
		{
			draw_poly_strf(32.0f, y, 616.0f, 1.0f, 1.0f, 1.0f, 1.0f, "%s: %u", options[counter].desc, options[counter].current);
		}
		else
		{
			draw_poly_strf(32.0f, y, 616.0f, 1.0f, 1.0f, 1.0f, 1.0f, "%s: %s", options[counter].desc, (options[counter].current ? "Yes": "No"));
		}
		if(counter == selected)
		{
			draw_poly_box(32.0f, y, 616.0f, y + 24, 5.2f, alpha, 1.0f, 0.0f, 1.0f, alpha, 1.0f, 0.0f, 1.0f);
		}
		y += 24;
		counter++;
	}
	alpha += dalpha;
	if(alpha <= 0.2f || alpha >= 1.0f)
	{
		dalpha *= -1;
	}
}

void options_runner()
{
	cont_cond_t cond, oldcond;
	uint8 times = 0;
	uint8 exit = 0;
	int current = 0;
	
	prep_image("/rd/border.jpg");

	while(!exit)
	{
		ta_begin_render();
		show_image(1.0f, PVR_LIST_OP_POLY);
		ta_commit_eol();

		draw_options(current);
		ta_commit_eol();
		ta_finish_frame();
		
		if(cont_get_cond(maple_first_controller(), &cond) < 0)
		{
			exit = 1;
		}
		else if(!(cond.buttons & CONT_DPAD_UP))
		{
			if((oldcond.buttons & CONT_DPAD_UP)||(times == 0))
			{
				if(current > 0) current--;
				times = 0;
			}
		}
		else if(!(cond.buttons & CONT_DPAD_DOWN))
		{
			if((oldcond.buttons & CONT_DPAD_DOWN)||(times == 0))
			{
				if(options[current+1].variable != NULL)
				{
					current++;
				}
				times = 0;
			}
		}
		else if(!(cond.buttons & CONT_A))
		{
			if(oldcond.buttons & CONT_A)
			{
				if(options[current].max > 1)
				{
					if(options[current].current < options[current].max)
					{
						options[current].current++;
					}
				}
				else
				{
					options[current].current = !options[current].current;
				}
			}
		}
		else if(!(cond.buttons & CONT_B))
		{
			if(oldcond.buttons & CONT_B)
			{
				if(options[current].max > 1)
				{
					if(options[current].current)
					{
						options[current].current--;
					}
				}
			}
		}
		else if(!(cond.buttons & CONT_START))
		{
			if(oldcond.buttons & CONT_START)
			{
				exit = 1;
			}
		}
		times++;
		times %= 12;
		oldcond = cond;
	}

	current = 0;
	while(options[current].variable != NULL)
	{
		*(options[current].variable) = options[current].current;
		current++;
	}
}


